from typing import Annotated, Literal, Optional, Tuple
import typing
from pydantic import BaseModel as PydanticBaseModel, ConfigDict, Field
from pydantic.alias_generators import to_camel


class BaseModel(PydanticBaseModel):
    model_config = ConfigDict(alias_generator=to_camel)


# data types


ZoneType = Literal["hardwired", "rf", "touchpad"]
ZoneStatus = Literal[
    "normal", "tripped", "faulted", "alarm", "trouble", "bypassed", "unknown"
]
PanelType = Literal["concord", "concordExpress", "concordExpress4", "concordEuro"]
ArmingLevel = Literal["zoneTest", "off", "home", "away", "night", "silent"]
Feature = Literal[
    "chime", "energySaver", "noDelay", "latchKey", "silentArm", "quickArm"
]


class ZoneData(BaseModel):
    partition_number: int
    area_number: int
    group_number: int
    zone_number: int
    zone_type: ZoneType
    zone_status: ZoneStatus
    zone_text: str

    def id(self) -> str:
        return f"p{self.partition_number}-z{self.zone_number}"

    def group_id(self) -> str:
        return f"p{self.partition_number}-g{self.group_number}"

    def callback_id(self) -> str:
        return self.id()


class ZoneStatusData(BaseModel):
    partition_number: int
    area_number: int
    zone_number: int
    zone_status: ZoneStatus

    def zone_id(self) -> str:
        return f"p{self.partition_number}-z{self.zone_number}"

    def callback_id(self) -> str:
        return self.zone_id()


class PartitionData(BaseModel):
    partition_number: int
    area_number: int
    arming_level: ArmingLevel
    zones: list[str]

    def id(self) -> int:
        return self.partition_number

    def callback_id(self) -> str:
        return f"p{self.partition_number}"


class GroupData(BaseModel):
    partition_number: int
    number: int
    zones: list[str]

    def id(self) -> str:
        return f"p{self.partition_number}-g{self.number}"


class PanelData(BaseModel):
    panel_type: PanelType
    hardware_revision: str
    software_revision: str
    serial_number: str


class ArmingLevelData(BaseModel):
    partition_number: int
    area_number: int
    arming_level: ArmingLevel

    def callback_id(self) -> str:
        return f"p{self.partition_number}"


class FeatureState(BaseModel):
    partition_number: int
    area_number: int
    feature_state: Feature


class TimeDate(BaseModel):
    hour: int
    minute: int
    month: int
    day: int
    year: int


class SuperBusDevData(BaseModel):
    partition_number: int
    area_number: int
    device_id: tuple[int, int, int]
    device_status: Literal["ok", "failed"]


class SuperBusCapabilityData(BaseModel):
    capability: str
    data: Optional[int] = None


class SuperBusDevCapData(BaseModel):
    device_id: tuple[int, int, int]
    capability: SuperBusCapabilityData


class SubEventData(BaseModel):
    sub_event: str
    user_number: Optional[tuple[int, int]] = None


class EventData(BaseModel):
    event: str
    data: SubEventData


class AlarmTroubleData(BaseModel):
    partition_number: int
    area_number: int
    source_type: str
    source_number: tuple[int, int, int]
    event: EventData


class TouchpadDisplay(BaseModel):
    partition_number: int
    area_number: int
    message_type: int
    display_tokens: list[int]
    text: str


Zones = dict[str, ZoneData]
Partitions = dict[int, PartitionData]
Groups = dict[str, GroupData]


class State(BaseModel):
    panel: PanelData
    zones: Zones
    partitions: Partitions
    groups: Groups


# message types


class ConcordAckMessage(BaseModel):
    type: Literal["ack"]


class ConcordNakMessage(BaseModel):
    type: Literal["nak"]


class ConcordPanelTypeMessage(BaseModel):
    type: Literal["panelType"]
    data: PanelData


class ConcordAutomationEventLostMessage(BaseModel):
    type: Literal["automationEventLost"]
    data: list[int]


class ZoneDataMessage(BaseModel):
    type: Literal["zoneData"]
    data: ZoneData


class PartitionDataMessage(BaseModel):
    type: Literal["partitionData"]
    data: PartitionData


class SuperBusDevDataMessage(BaseModel):
    type: Literal["superBusDevData"]
    data: SuperBusDevData


class SuperBusDevCapMessage(BaseModel):
    type: Literal["superBusDevCap"]
    data: SuperBusDevCapData


class OutputDataMessage(BaseModel):
    type: Literal["outputData"]
    data: list[int]


class EqptListDoneMessage(BaseModel):
    type: Literal["eqptListDone"]


class SchedDataMessage(BaseModel):
    type: Literal["schedData"]
    data: list[int]


class SchedEventDataMessage(BaseModel):
    type: Literal["schedEventData"]
    data: list[int]


class LightAttachMessage(BaseModel):
    type: Literal["lightAttach"]
    data: list[int]


class ClearImageMessage(BaseModel):
    type: Literal["clearImage"]
    data: list[int]


class ZoneStatusMessage(BaseModel):
    type: Literal["zoneStatus"]
    data: ZoneStatusData


class ArmingLevelMessage(BaseModel):
    type: Literal["armingLevel"]
    data: ArmingLevelData


class AlarmTroubleMessage(BaseModel):
    type: Literal["alarmTrouble"]
    data: AlarmTroubleData


class EntryExitDelayMessage(BaseModel):
    type: Literal["entryExitDelay"]
    data: list[int]


class SirenSetupMessage(BaseModel):
    type: Literal["sirenSetup"]
    data: list[int]


class SirenSyncMessage(BaseModel):
    type: Literal["sirenSync"]


class SirenGoMessage(BaseModel):
    type: Literal["sirenGo"]


class TouchpadMessage(BaseModel):
    type: Literal["touchpad"]
    data: TouchpadDisplay


class SirenStopMessage(BaseModel):
    type: Literal["sirenStop"]
    data: list[int]


class FeatStateMessage(BaseModel):
    type: Literal["featState"]
    data: FeatureState


class TempMessage(BaseModel):
    type: Literal["temp"]
    data: list[int]


class TimeAndDateMessage(BaseModel):
    type: Literal["timeAndDate"]
    data: TimeDate


class LightsStateMessage(BaseModel):
    type: Literal["lightsState"]
    data: list[int]


class UserLightsMessage(BaseModel):
    type: Literal["userLights"]
    data: list[int]


class KeyfobMessage(BaseModel):
    type: Literal["keyfob"]
    data: list[int]


ConcordReceivedMessageType = (
    ConcordAckMessage
    | ConcordNakMessage
    | ConcordPanelTypeMessage
    | ConcordAutomationEventLostMessage
    | ZoneDataMessage
    | PartitionDataMessage
    | SuperBusDevDataMessage
    | SuperBusDevCapMessage
    | OutputDataMessage
    | EqptListDoneMessage
    | SchedDataMessage
    | SchedEventDataMessage
    | LightAttachMessage
    | ClearImageMessage
    | ZoneStatusMessage
    | ArmingLevelMessage
    | AlarmTroubleMessage
    | EntryExitDelayMessage
    | SirenSetupMessage
    | SirenSyncMessage
    | SirenGoMessage
    | TouchpadMessage
    | SirenStopMessage
    | FeatStateMessage
    | TempMessage
    | TimeAndDateMessage
    | LightsStateMessage
    | UserLightsMessage
    | KeyfobMessage
)


class ConcordReceivedMessage(BaseModel):
    type: Literal["message"]
    data: ConcordReceivedMessageType = Field(..., discriminator="type")


class StateReceivedMessage(BaseModel):
    type: Literal["state"]
    data: State


ConcordMessageType = StateReceivedMessage | ConcordReceivedMessage
ReceivableMessage = Annotated[ConcordMessageType, Field(discriminator="type")]


# command types

ListRequestType = Literal[
    "allData",
    "zoneData",
    "partData",
    "busDevData",
    "busCapData",
    "outputData",
    "scheduleData",
    "eventData",
    "lightAttach",
]

Keypress = Literal[
    "0",
    "1",
    "2",
    "3",
    "4",
    "5",
    "6",
    "7",
    "8",
    "9",
    "*",
    "#",
    "policePanic",
    "auxPanic",
    "firePanic",
    "lightsOn",
    "lightsOff",
    "lightsToggle",
    "keyswitchOn",
    "keyswitchOff",
    "keyswitchToggle",
    "fireTPAcknowledge",
    "fireTPSilence",
    "fireTPFireTest",
    "fireTPSmokeReset",
    "keyfobDisarm",
    "keyfobArm",
    "keyfobLights",
    "keyfobStar",
    "keyfobArmDisarm",
    "keyfobLightsStar",
    "keyfobLongLights",
    "keyfobDirectArmLevelThree",
    "keyfobDirectArmLevelTwo",
    "keyfobArmStar",
    "keyfobDisarmLights",
    "TPAKey",
    "TPBKey",
    "TPCKey",
    "TPDKey",
    "TPEKey",
    "TPFKey",
]

ArmMode = Literal["stay", "away"]
ArmLevel = Literal["normal", "silent", "instant"]


class ConcordListCommand(BaseModel):
    message: Literal["list"] = "list"
    params: ListRequestType


class ArmOptions(BaseModel):
    mode: ArmMode
    code: list[Keypress]
    level: Optional[ArmLevel] = None
    partition: Optional[int] = None


class DisarmOptions(BaseModel):
    code: list[Keypress]
    partition: Optional[int] = None


class ConcordArmCommand(BaseModel):
    message: Literal["arm"] = "arm"
    params: ArmOptions


class ConcordDisarmCommand(BaseModel):
    message: Literal["disarm"] = "disarm"
    params: DisarmOptions


class ConcordToggleChimeCommand(BaseModel):
    message: Literal["toggleChime"] = "toggleChime"
    params: Optional[int] = None


class ConcordKeypressCommand(BaseModel):
    message: Literal["keypress"] = "keypress"
    params: Tuple[int, list[Keypress]]


class ConcordDynamicDataRefreshCommand(BaseModel):
    message: Literal["dynamicDataRefresh"] = "dynamicDataRefresh"


ConcordCommandType = (
    ConcordListCommand
    | ConcordArmCommand
    | ConcordDisarmCommand
    | ConcordToggleChimeCommand
    | ConcordKeypressCommand
    | ConcordDynamicDataRefreshCommand
)


class CommandSendableMessage(BaseModel):
    type: Literal["command"] = "command"
    data: ConcordCommandType = Field(..., discriminator="message")


class GetStateSendableMessage(BaseModel):
    type: Literal["getState"] = "getState"


SendableMessageType = CommandSendableMessage | GetStateSendableMessage
SendableMessage = Annotated[SendableMessageType, Field(discriminator="type")]


# helper functions
def code_to_keypresses(code: str) -> list[Keypress]:
    all_keypresses = typing.get_args(Keypress)

    try:
        return [all_keypresses[all_keypresses.index(char)] for char in code]
    except ValueError:
        raise Concord4KeypressError("invalid keypress in code")


class Concord4KeypressError(Exception):
    """Class for exceptions for Concord4's keypress"""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)



#!/usr/bin/env python

import asyncio
import signal
from typing import Callable, Optional
from concord4ws.types import (
    ArmLevel,
    ArmMode,
    ArmOptions,
    ArmingLevelData,
    CommandSendableMessage,
    ConcordArmCommand,
    ConcordDisarmCommand,
    ConcordToggleChimeCommand,
    DisarmOptions,
    Keypress,
    PanelData,
    PartitionData,
    ReceivableMessage,
    SendableMessage,
    State,
    ZoneData,
    ZoneStatusData,
)
from pydantic import TypeAdapter
import websockets
import logging

logger = logging.getLogger("concord4ws")


class Concord4WSClient:
    _connected: bool = False
    _connect_attempt: int = 0
    _callbacks = dict[str, set[Callable[[], None]]]()
    _initialized: bool = False
    _state: State
    _ws: websockets.WebSocketClientProtocol

    def __init__(self, host: str, port: str | int):
        self.host = host
        self.port = port
        self._loop = asyncio.get_event_loop()

    @property
    def state(self):
        return self._state

    @property
    def connected(self):
        return self._connected

    async def test_connect(self) -> bool:
        try:
            async with websockets.connect(f"ws://{self.host}:{self.port}"):
                return True
        except Exception as _:
            return False

    async def _handle_signal_exit(self):
        logger.info("received signal, closing connection")
        if self._ws is not None and not self._ws.closed:
            await self._ws.close()

        self._connected = False

        self._loop.stop()
        self._loop.close()

    async def _event_loop(self):
        self._loop.add_signal_handler(
            signal.SIGTERM, lambda: asyncio.create_task(self._handle_signal_exit())
        )

        async for ws in websockets.connect(f"ws://{self.host}:{self.port}"):
            try:
                self._connected = True
                self._connect_attempt = 0
                self._ws = ws

                while True:
                    async for message in ws:
                        logger.debug("message received: %s", message)
                        self._handle_message(message)

            except websockets.ConnectionClosed:
                self._connected = False
                self._connect_attempt += 1

                continue

    def _handle_message(self, message: websockets.Data) -> None:
        """Handle incoming message from server."""
        try:
            recv: ReceivableMessage = TypeAdapter(ReceivableMessage).validate_json(
                message
            )
            logger.debug("decoded message as: %s", recv)

            if recv.type == "state":
                logger.debug("updating state")
                self._state = recv.data
                self._initialized = True

            if recv.type == "message":
                match recv.data.type:
                    case "panelType":
                        logger.debug("handling panel type")
                        self._handle_panel_type(recv.data.data)
                    case "zoneData":
                        logger.debug("handling zone data")
                        self._handle_zone_data(recv.data.data)
                    case "zoneStatus":
                        logger.debug("handling zone status data")
                        self._handle_zone_status_data(recv.data.data)
                    case "partitionData":
                        logger.debug("handling partition data")
                        self._handle_partition_data(recv.data.data)
                    case "armingLevel":
                        logger.debug("handling arming level data")
                        self._handle_arming_level_data(recv.data.data)
                    case _:
                        logger.debug("unhandled message type")

        except Exception as e:
            logger.error(e)

    def _handle_panel_type(self, data: PanelData) -> None:
        """Handle panel type message."""
        self._state.panel = data

    def _handle_zone_data(self, data: ZoneData) -> None:
        """Handle zone data message."""
        self._state.zones[data.id()] = data

        if data.callback_id() in self._callbacks:
            for callback in self._callbacks[data.id()]:
                callback()

    def _handle_zone_status_data(self, data: ZoneStatusData) -> None:
        """Handle zone status data message."""
        self._state.zones[data.zone_id()].zone_status = data.zone_status

        if data.callback_id() in self._callbacks:
            for callback in self._callbacks[data.callback_id()]:
                callback()

    def _handle_partition_data(self, data: PartitionData) -> None:
        """Handle partition data message."""
        self._state.partitions[data.id()] = data

        if data.callback_id() in self._callbacks:
            for callback in self._callbacks[data.callback_id()]:
                callback()

    def _handle_arming_level_data(self, data: ArmingLevelData) -> None:
        """Handle arming level data message."""
        self._state.partitions[data.partition_number].arming_level = data.arming_level

        if data.callback_id() in self._callbacks:
            for callback in self._callbacks[data.callback_id()]:
                callback()

    def register_callback(
        self, zone_or_part_id: str, callback: Callable[[], None]
    ) -> None:
        """Register callback, called when zone or partition specified changes state."""
        logger.debug("registering callback for: %s", zone_or_part_id)
        self._callbacks.setdefault(zone_or_part_id, set()).add(callback)

    def remove_callback(
        self, zone_or_part_id: str, callback: Callable[[], None]
    ) -> None:
        """Remove previously registered callback."""
        logger.debug("removing callback for: %s", zone_or_part_id)
        self._callbacks[zone_or_part_id].remove(callback)

    async def connect(self) -> None:
        """Connect to Concord4WS."""
        logger.info("connecting to concord4ws server")

        self._loop.create_task(self._event_loop())

        # wait for first message from server to initialize state
        while not self._initialized:
            await asyncio.sleep(0.1)

        logger.info("connected to concord4ws server")

    async def disconnect(self) -> None:
        """Disconnect from Concord4WS."""
        if not self._connected:
            logger.error("cannot disconnect, not connected")
            return

        logger.info("disconnecting from concord4ws server")

        self._loop.stop()
        self._loop.close()
        self._loop = asyncio.get_event_loop()

        self._connected = False

    async def send(self, message: SendableMessage) -> None:
        """Send message to server."""
        if not self._connected:
            logger.error("cannot send message, not connected")
            raise Concord4WSClientError("not connected to server")

        logger.debug("sending message: %s", message)

        await self._ws.send(message.model_dump_json())

    async def arm(
        self,
        mode: ArmMode,
        code: list[Keypress],
        level: Optional[ArmLevel] = None,
        partition: Optional[int] = None,
    ) -> None:
        """Arm partition."""

        await self.send(
            CommandSendableMessage(
                data=ConcordArmCommand(
                    params=ArmOptions(
                        mode=mode, code=code, level=level, partition=partition
                    )
                )
            )
        )

    async def disarm(
        self,
        code: list[Keypress],
        partition: Optional[int] = None,
    ) -> None:
        """Disarm partition."""

        await self.send(
            CommandSendableMessage(
                data=ConcordDisarmCommand(
                    params=DisarmOptions(code=code, partition=partition)
                )
            )
        )

    async def toggle_chime(self, partition: Optional[int]) -> None:
        """Toggle chime for partition."""

        await self.send(
            CommandSendableMessage(data=ConcordToggleChimeCommand(params=partition))
        )


class Concord4WSClientError(Exception):
    """Class for exceptions in Concord4WSClient."""

    def __init__(self, message: str):
        self.message = message
        super().__init__(self.message)
